#include <Arduino.h>
#include <Crypto.h>
#include <DFRobot_ID809.h>
#include <EEPROM.h>

// use serial3 as interface with fingerprint
// sensor.

#define FPSerial Serial3

// define buillt_in led

#define BUILTIN_LED 13

// Collect the fingerprint 3 times

#define COLLECT_NUMBER 3

/*------Global variables--------------*/

DFRobot_ID809 fingerprint;
uint8_t ID, i, ret;
byte hash[SHA256_SIZE];
uint8_t temp[1008];
// EEPROM OFFSET to store

const uint16_t EEPROM_FP_HASH = 1696;

/*---------Compute hash----------------*/

void compute_hash() {

  SHA256 hasher;
  hasher.doUpdate(temp, sizeof(temp));
  hasher.doFinal(hash);

  Serial.println("-----------------------");
  for (uint8_t i = 0; i < SHA256_SIZE; i++) {

    if (hash[i] < 0x10) {

      Serial.print('0');
    }

    Serial.print(hash[i], HEX);
    Serial.print(" ");
  }
  Serial.println("-----------------------");
}

/*-----Blink on board led -------------- */

void blink_led() {
  digitalWrite(BUILTIN_LED, HIGH);
  delay(500);
  digitalWrite(BUILTIN_LED, LOW);
  delay(500);
}

/*-------- Save FingerPrint hash in EEPROM ------*/

void compareHash() {
  uint16_t memoryAddress;
  byte stored_hash[SHA256_SIZE];

  memoryAddress = EEPROM_FP_HASH + ((ID - 1) * 32);

  Serial.println("---------------------------");
  for (uint8_t i = 0; i < SHA256_SIZE; i++) {

    stored_hash[i] = EEPROM.read(memoryAddress);
    Serial.print(stored_hash[i], HEX);
    Serial.print(" ");
    memoryAddress++;
  }
  Serial.println("--------------------------");

  for (uint8_t i = 0; i < SHA256_SIZE; i++) {

    if (stored_hash[i] != hash[i]) {
      Serial.println("Hash not matched");
      return;
    }
  }
  Serial.println("Hash matched");
}

/*---------Arduino setup function------*/

void setup() {
  // config led_pin as output
  pinMode(BUILTIN_LED, OUTPUT);
  // use serial port to print the baud rate
  Serial.begin(9600);
  // config fingerprint sensor
  FPSerial.begin(115200);
  fingerprint.begin(FPSerial);

  while (fingerprint.isConnected() == false) {
    Serial.println("Communication with the device failed");
    delay(1000);
  }
}

/*---------Register Finger print ------------*/

void matchFingerprint() {

  /*Set fingerprint LED ring mode, color, and number of blinks
    Can be set as follows:
    Parameter 1:<LEDMode>
    eBreathing   eFastBlink   eKeepsOn    eNormalClose
    eFadeIn      eFadeOut     eSlowBlink
    Parameter 2:<LEDColor>
    eLEDGreen  eLEDRed      eLEDYellow   eLEDBlue
    eLEDCyan   eLEDMagenta  eLEDWhite
    Parameter 3:<Number of blinks> 0 represents blinking all the time
    This parameter will only be valid in mode eBreathing, eFastBlink,
    eSlowBlink
   */
  fingerprint.ctrlLED(/*LEDMode = */ fingerprint.eBreathing,
                      /*LEDColor = */ fingerprint.eLEDBlue,
                      /*blinkCount = */ 0);

  Serial.println("Plase press down your finger");

  if ((fingerprint.collectionFingerprint(/*timeout = */ 0)) != ERR_ID809) {
    /*Set fingerprint LED ring to quick blink in yellow 3 times */
    fingerprint.ctrlLED(/*LEDMode = */ fingerprint.eFastBlink,
                        /*LEDColor = */ fingerprint.eLEDYellow,
                        /*blinkCount = */ 3);
    Serial.println("capturing succeeds");
    Serial.println("Please release your finger");

    while (fingerprint.detectFinger())
      ;

    /*Compare the captured fingerprint with all the fingerprints
     * inthe fingerprint library return fingerprint ID if succeed,
     * return 0 when failed
     * */
    ret = fingerprint.search();

    if (ret != 0) {

      /*Set fingerprint LED ring to always ON in green */
      fingerprint.ctrlLED(/*LEDMode = */ fingerprint.eKeepsOn,
                          /*LEDColor = */ fingerprint.eLEDGreen,
                          /*blinkCount = */ 0);
      Serial.print("Matching succeeds,ID=");
      ID = ret;
      Serial.println(ret);
      fingerprint.getTemplate(ID, temp);
      delay(100);
      compute_hash();
      compareHash();

    } else {
      /*Set fingerprint LED ring to always ON in red*/
      fingerprint.ctrlLED(/*LEDMode = */ fingerprint.eKeepsOn,
                          /*LEDColor = */ fingerprint.eLEDRed,
                          /*blinkCount = */ 0);
      Serial.println("Matching fails");
    }
  } else {
    Serial.println("Capturing failes");
    /*Get error code information*/
    // desc = fingerprint.getErrorDescription();
    // Serial.println(desc);
  }
}

/*-----------Arduino loop function--------- */

void loop() {
  blink_led();
  matchFingerprint();
}
